#!/usr/local/bin/ruby

require 'cgi'
require 'erb'
require 'delegate'
require 'enumerator'
require 'iconv'
require 'net/http'
require 'open-uri'
require 'singleton'
require 'uri'

require 'cocoron'


module Enumerable
   def group_by(&block)
      inject(Hash.new){|result, elem|
	 (result[yield(elem)] ||= []) << elem
	 result
      }.values
   end
end


class CGI
   def get_param(name)
      unify_encoding(self.params.fetch(name).first)
   rescue IndexError
      if block_given? then
	 yield
      else
	 raise
      end
   end

   def get_params(name)
      self.params.fetch(name){ Array.new }.collect{|s| unify_encoding(s) }
   end
end


class Preference < DelegateClass(Hash)
   include Singleton

   def initialize
      @pref = File.open('./cocoronrc'){|f| YAML.load(f) }
      @pref['lang'] ||= 'en'
      @pref['libs'] ||= %w(books)
      @pref['libpath'] ||= ENV['HOME']
      @pref['title'] ||= 'Cocoron'
      @pref['script-path'] ||= nil

      untaint_rec(@pref)
      super(@pref)
   end

private
   def untaint_rec(obj)
      case obj
      when Array then
	 obj.each{|elem| untaint_rec(elem) }
      when Hash then
	 obj.each_value{|elem| untaint_rec(elem) }
      end

      obj.untaint
   end
end



if Preference.instance.has_key?('http-proxy') then
   ENV['http_proxy'] = Preference.instance['http-proxy']
else
   $SAFE = 1
end



begin
   require 'gettext'
   include GetText
   bindtextdomain('cocoron', Preference.instance['locale-path'],
                  Preference.instance['lang'], 'utf-8')
rescue LoadError then
   def _(str)
      str.gsub(/([a-zA-Z])_([a-zA-Z])/){ "#{$1} #{$2}" }
   end

   def n_(s, t, n)
      _(n == 1 ? s : t)
   end
end



def unify_encoding(str)
   self.send("unify_encoding_#{Preference.instance['lang']}", 'UTF-8', str)
end


# do nothing for now
def unify_encoding_en(to, str)
   str
end


def unify_encoding_ja(to, str)
   if /\e(?:\(B|\(J|\$@|\$B)/ === str then
      Iconv.conv(to, 'ISO-2022-JP', str)
   else
      begin
         Iconv.conv(to, 'UTF-8', str)
      rescue Iconv::Failure
         begin
            Iconv.conv(to, 'EUC-JP', str)
         rescue Iconv::Failure
            Iconv.conv(to, 'SHIFT_JIS', str)
         end
      end
   end
end


def libpaths
   Preference.instance['libs'].collect{|e| libname2path(e) }
end


def libname2path(name)
   if n = Preference.instance['libs'].detect{|l| l == name } then
      File.join(Preference.instance['libpath'], n)
   else
      nil
   end
end


def load_library(name, mode = :r)
   path = libname2path(name)

   if libpaths.include?(path) then
      Cocoron::Library.new(path, mode)
   else
      raise IndexError
   end
end



class Page
   include ERB::Util

   def initialize(cgi)
      @cgi = nil
      raise 'Subclass Responsibility'
   end

   def show
      if invoke_from_update_cgi? then
	 @cgi.out({'status' => 'MOVED',
		   'location' => self.url,
		   'type' => 'text/plain'
		  }) { "301 Moved. please access #{self.url}" }
      else
	 @cgi.out({'charset' => 'utf-8',
		   'type' => self.content_type
		  }){ self.to_s }
      end
   end

   def to_s
      raise 'Subclass Responsibility'
   end

   def url
      raise 'Subclass Responsibility'
   end

   def accesskey
      raise 'Subclass Responsibility'
   end

   def content_type
      'text/html'
   end

private
   def invoke_from_update_cgi?
      not @cgi.script_name.nil? and
         File.basename(@cgi.script_name) != File.basename(index_url) and
	 self.respond_to?(:url)
   end

   def index_url
      script_path + 'index.cgi'
   end

   def update_url
      script_path + 'update.cgi'
   end

   def script_path
      Preference.instance['script-path'] ||
         ('http://%s%s/' % [@cgi.server_name, File.dirname(@cgi.script_name)])
   end

   def image_cache_path(original_url)
      File.join('images', File.basename(URI.unescape(URI.parse(original_url.untaint).path)))
   end

   def image_cache_url(original_url)
      if not original_url or original_url.empty? then
	 nil
      else
	 path = image_cache_path(original_url)

	 File.exist?(path) ? path : nil
      end
   rescue OpenURI::HTTPError, SocketError, Timeout::Error
      nil
   end

   GIF_MAGICNUM = 'GIF8'

   def gif?(str)
      str[0, (GIF_MAGICNUM.length)] == GIF_MAGICNUM
   end
   
   def cache_image!(book)
      nbook = nil

      %w(small medium large).each do |size|
         attr = "image_url_#{size}"

         if book.send(attr).empty? then
            nbook ||= book.id.create_book()
            book.send("#{attr}=", nbook.send(attr).untaint)
         end

	 unless book.send(attr).empty? then
	    uri = book.send(attr)
	    path = image_cache_path(uri)
	    
	    unless File.exist?(path) then
	       open(uri) do |f|
                  cont = f.read

                  gif?(cont) or open(path, 'w'){|out| out.write(cont) }
               end
	    end
	 end
      end
   end

   def authors_to_s(as)
      as.nil? ? '' : as.join(_(', '))
   end

   def query_str(path, params)
      "#{path}?" + params.collect{|k, v|
	              "#{CGI.escape(k.to_s)}=#{CGI.escape(v.to_s)}" }.join('&')
   end
end


class HTMLPage < Page
   def to_s
      lang = Preference.instance['lang']

      <<-EOS
<?xml version="1.0" encoding="UTF-8" ?>
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"
    "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">

<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="#{lang}">
<head>
    <link rel="stylesheet" type="text/css" href="./cocoron.css" />
    <link rel="alternate" type="application/rss+xml" href="#{RSSPage.new(@cgi).url}" />
    <script type="text/javascript" src="./js/util.js"></script>
    <meta name="robots" content="noindex,nofollow" />
    <title>#{self.title}</title>
</head>
<body id="#{self.body_id}">
#{self.navigation}<div id="content">
    <h1>#{self.title}</h1>
#{self.body}</div>
</body>
</html>
      EOS
   end

   def title
      raise 'Subclass Responsibility'
   end

   def body
      raise 'Subclass Responsibility'
   end

   def navigation
      ''
   end

   def body_id
      self.class.name
   end
end


class RSSPage < Page
   N_BOOKS = 20

   def initialize(cgi, mode = 'ctime')
      require 'rss/maker'

      @cgi = cgi
      @mode = mode
   end

   def to_s
      lib_and_books = libpaths.collect{|lib_path|
			 lib = Cocoron::Library.new(lib_path)
		         lib.values \
			    .sort_by{|b| b.send(@mode) } \
			    .reverse \
			    .first(N_BOOKS) \
			    .collect{|b| [lib, b] }
		      }.inject([]){|res, books| res.concat(books) } \
		       .sort_by{|(lib, book)| book.send(@mode) } \
		       .reverse \
		       .first(N_BOOKS)

      r = RSS::Maker.make('1.0'){|maker|
	 maker.channel.about = self.url
	 maker.channel.title = Preference.instance['title']
	 maker.channel.link = index_url
	 maker.channel.description = ''

	 lib_and_books.each do |(lib, book)|
	    item = maker.items.new_item

	    item.title = '[%s] %s' %
	                    [lib.name, book.title]
	    item.date = book.send(@mode)
	    item.link = ViewPage.new(@cgi, lib, book).url
	    item.description = book.note
	 end
      }.to_s
   end

   def content_type
      'application/rss+xml'
   end

   def url
      query_str(index_url,
                {:cmd => :rss,
		 :mode => @mode
		})
   end
end

class RDFPage < Page
   def initialize(cgi, id_str)
      @cgi = cgi
      id = begin
	      id_str && ::Cocoron::BookID.parse(id_str)
	   rescue ::Cocoron::BookID::InvalidBookIDError then
	      nil
	   end
      @book = id && libpaths.collect{|l| ::Cocoron::Library.new(l)[id.code] }.detect{|e| e }
   end

   def show
      if @book.nil? or not %w(ISBN ISSN).include?(@book.id.sort) then
	 @cgi.out({'status' => 'NOT_FOUND',
		   'type' => 'text/plain'
		  }){ '404 Not Found.' }
      else
	 super
      end
   end

   def content_type
      'application/xml'
   end

   def to_s
      erb = ERB.new(<<-EOS, nil, '>')
<rdf:RDF
  xmlns:rdf="http://www.w3c.org/1999/02/22-rdf-syntax-ns#"
  xmlns:dc="http://purl.org/dc/elements/1.1/">
    <rdf:Description rdf:about="urn:<%=h @book.id.sort %>:<%=h @book.id.code %>">
       <dc:title><%=h @book.title %></dc:title>
<% @book.authors.each do |a| %>
       <dc:creator><%=h a %></dc:creator>
<% end %>
       <dc:subject><%=h @book.edition %></dc:subject>
       <dc:publisher><%=h @book.publisher %></dc:publisher>
       <dc:date><%=h @book.release_date.gsub('/', '-') %></dc:date>
    </rdf:Description>
</rdf:RDF>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end
end


class ListPage < HTMLPage
   @@npages = 8
   @@nbooks = 15

   def initialize(cgi, library, start_index = 0, nbook = self.class.nbooks, sort = nil, marker = nil)
      @cgi = cgi
      @library = library
      @start_index = nbook.zero? ? 0 : start_index
      @nbook = nbook
      @marker = marker
      @sort = sort
      @accesskey = '.'
   end

   def self.contains(book, library, cgi, sort = nil)
      n = @@nbooks
      i = library.keys.index(book.id.code)

      s = i.nil? ? 0 : n * i.div(n)

      self.new(cgi, library, s, n, nil, book.id.code)
   end

   def self.nbooks
      @@nbooks
   end

   def self.nbooks=(n)
      @@npages = n
   end

   def self.npages
      @@npages
   end

   def self.npages=(n)
      @@npages = n
   end

   def url
      query_str(index_url,
                {:cmd => :list,
                 :lib => @library.name,
                 :n => @nbook, :s => @start_index,
		 :sort => @sort}) +
	 (@marker ? '#id-' + @marker : '')
   end

   def prev_page
      if @nbook <= 0 or @start_index <= 0 then
         nil
      else
         s = [0, @start_index - @nbook].max
         pp = self.class.new(@cgi, @library, s, @nbook, @sort)
	 pp.instance_eval{ @accesskey = ',' }
	 pp
      end
   end

   def next_page
      if @nbook <= 0 or @start_index + @nbook >= @library.size then
         nil
      else
         s = @start_index + @nbook
         np = self.class.new(@cgi, @library, s, @nbook, @sort)
	 np.instance_eval{ @accesskey = ';' }
	 np
      end
   end

   def body
      erb = ERB.new(<<-EOS, nil, '>')
   <table border="1" summary="<%=h self.title %>">
   <thead>
      <tr>
         <td></td>
         <th abbr="<%=h _('ID') %>" <%= class_of_th('id') %>>
             <%= th_label('id', 'ID') %>
	 </th>
         <th abbr="<%=h _('title') %>" <%= class_of_th('title') %>>
             <%= th_label('title', 'title') %>
	 </th>
         <th abbr="<%=h _('authors') %>" <%= class_of_th('authors') %>>
             <%= th_label('authors', 'authors') %>
	 </th>
         <th abbr="<%=h _('publisher') %>" <%= class_of_th('publisher') %>>
             <%= th_label('publisher', 'publisher') %>
	 </th>
         <th abbr="<%=h _('edition') %>" <%= class_of_th('edition') %>>
             <%= th_label('edition', 'edition') %>
	 </th>
         <th abbr="<%=h _('release_date') %>" <%= class_of_th('release_date') %>>
             <%= th_label('release_date', 'release_date') %>
	 </th>
         <th abbr="<%=h _('price') %>" <%= class_of_th('price') %>>
             <%= th_label('price', 'price') %>
	 </th>
         <th abbr="<%=h _('note') %>" <%= class_of_th('note') %>>
             <%= th_label('note', 'note') %>
	 </th>
      </tr>
   </thead>
   <tbody>
<%
         books_to_show.each do |book|
            img_s = image_cache_url(book.image_url_small)
            img_l = image_cache_url(book.image_url_large)
            vpage = ViewPage.new(@cgi, @library, book)
%>
      <tr id="id-<%=h book.id.code %>">
         <td class="cover"><%# %>
<%          if img_l then %><a href="<%=h img_l %>"><% end %>
<%          if img_s then %><img src="<%=h img_s %>" alt="o" <%# %>
<%#%>title="<%=h _("Cover page of `%s'") % [book.title] %>" /><% end %>
<%          if img_l then %></a><% end %>
<%#    %></td>
         <td><a href="<%=h vpage.url %>"><%=h book.id.code %></a></td>
         <td class="title"><%=h book.title %></td>
         <td class="authors"><%=h authors_to_s(book.authors) %></td>
         <td class="publisher"><%=h book.publisher %></td>
         <td class="edition"><%=h book.edition %></td>
         <td class="release_date"><%=h book.release_date %></td>
         <td class="price"><%=h book.price %></td>
         <td class="note"><pre><%=h book.note %></pre></td>
      </tr>
<%
         end
%>
   </tbody>
   </table>

<%= footer %>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def title
      _('Book list of %s') % [@library.name] +
         if @nbook > 0 then
            _("(%d - %d / %d)") %
               [@start_index + 1,
                [@start_index + @nbook, @library.size].min,
                @library.size]
         else
            ''
         end
   end

   def navigation
      tp = TopPage.new(@cgi)
      cp = CoverListPage.new(@cgi, @library, @start_index)
      sp = SearchPage.new(@cgi)
      pp = self.prev_page
      np = self.next_page

      erb = ERB.new(<<-EOS, nil, '>')
   <div id="navigation">
      <a href="<%=h tp.url %>" accesskey="<%=h tp.accesskey %>"><%=h _('top') %></a>
      <a href="<%=h cp.url %>" accesskey="<%=h cp.accesskey %>"><%=h _('cover list') %></a>
      <a href="<%=h sp.url %>" accesskey="<%=h sp.accesskey %>"><%=h _('search') %></a>
<%	if pp then %>
      <a href="<%=h pp.url %>" accesskey="<%=h pp.accesskey %>">
<%	end %>
	  <%=h _('prev') %>
<%	if pp then %>
      </a>
<%	end %>
<%	if np then %>
      <a href="<%=h np.url %>" accesskey="<%=h np.accesskey %>">
<%	end %>
	  <%=h _('next') %>
<%	if np then %>
      </a>
<%	end %>
      <a href="<%=h ListPage.new(@cgi, @library, 0, @nbook, 'ctime:r').url %>"><%=h _('recently added') %></a>
      <a href="<%=h ListPage.new(@cgi, @library, 0, @nbook, 'mtime:r').url %>"><%=h _('recently changed') %></a>
   </div>
   <form id="books-per-page" action="<%=h index_url %>" method="get">
	 <label><select name="n">
<% if @nbook.zero? then %>
		    <option value="0" selected="selected"><%=h _('all') %></option>
<% else %>
		    <option selected="selected"><%=h @nbook %></option>
<% end %>
<% [15, 30, 50, 100].each do |n| %>
<%    unless n == @nbook then %>
		    <option><%=h n %></option>
<%    end %>
<% end %>
<% unless @nbook.zero? then %>
		    <option value="0"><%=h _('all') %></option>
<% end %>
		</select>
		<%=h _('books/page') %></label>
         <input type="hidden" name="lib" value="<%=h @library.name %>" />
         <input type="hidden" name="s" value="<%=h @start_index %>" />
         <input type="hidden" name="sort" value="<%=h @sort %>" />
         <input type="hidden" name="cmd" value="list" />
         <input type="submit" value="<%=h _('show') %>" />
   </form>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def footer
      erb = ERB.new(<<-EOS, nil, '>')
   <div id="pages">
<% footer_links.each do |(text, page)| %>
<%    if page then %>
      <a href="<%=h page.url %>"><%=h text %></a>
<%    else %>
      <span><%=h text %></span>
<%    end %>
<% end %>
   </div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def footer_links
      result = Array.new
      unless @nbook.zero? then
	 npages_before = self.class.npages.div(2)
	 npages_after = self.class.npages - npages_before - 1
	 sind = @start_index.div(@nbook)
	 page_nums = 0.enum_for(:step, @library.size - 1, @nbook).to_a

	 if page_nums.size - 1 < sind + npages_after then
	    lb = [0, sind - self.class.npages].max
	    ub = page_nums.size - 1
	 elsif sind - npages_before < 0 then 
	    lb = 0 
	    ub = [self.class.npages, page_nums.size].min - 1
	 else 
	    lb = sind - npages_before
	    ub = sind + npages_after
	 end
      end

      return result if @nbook.zero? or lb == ub

      if lb != 0 then
	 first_page = self.class.new(@cgi, @library, page_nums.first, @nbook, @sort)
	 result << ["1-#{@nbook}", first_page]
      end

      if lb > 1 then
	 result << ['...', nil]
      end

      lb.upto(ub) do |i|
	 page = (page_nums[i] == @start_index ?
		    nil :
		    self.class.new(@cgi, @library, page_nums[i], @nbook, @sort))
	 result << ["#{page_nums[i] + 1}-#{[page_nums[i] + @nbook, @library.size].min}",
		    page]
      end

      if ub < page_nums.size - 2 then
	 result << ['...', nil]
      end

      if ub != page_nums.size - 1 then
	 last_page = self.class.new(@cgi, @library, page_nums.last, @nbook, @sort)
	 result << ["#{page_nums.last + 1}-#{@library.size}", last_page]
      end

      result
   end

   def accesskey
      @accesskey || super	# avoid to forget initializing
   end

private
   def books_to_show
      if @start_index > @library.size then
	 []
      elsif @nbook > 0 then
	 if sort_key_provided? then
	    sorted_books[@start_index, @nbook]
	 else
	    keys = @library.keys[@start_index, @nbook]
	    @library.values_at(*keys)
	 end
      else
	 if sort_key_provided? then
	    sorted_books
	 else
	    @library.values
	 end
      end
   end

   def sort_key_provided?
      (@sort and not @sort.empty?) &&
	 ((like_rev_key? && sort_key?(@sort[0..-3])) ||
	     sort_key?(@sort))
   end

   def sort_key?(str)
      Cocoron::Book.method_defined?(str)
   end

   def like_rev_key?
      /:r$/ === @sort
   end

   def sort_key
      [like_rev_key?, @sort.sub(/:r$/, '')]
   end

   def class_of_th(attr)
      if @sort == attr then
         'class="key-for-ascending-order"'
      elsif @sort == ('%s:r' % attr) then
         'class="key-for-descending-order"'
      else
         ''	# nothing
      end
   end

   def th_label(attr, label)
      '<a href="%s">%s</a>' %
         [@sort == attr ?
          ListPage.new(@cgi, @library, 0, @nbook, '%s:r' % attr).url :
          ListPage.new(@cgi, @library, 0, @nbook, attr).url,
          _(label)].collect{|str| CGI.escapeHTML(str) }
   end

   def sorted_books
      revp, key = sort_key

      result = @library.values.sort_by{|b| b.send(key) }
      result.reverse! if revp
      result
   end
end


class CoverListPage < ListPage
   def initialize(cgi, library, start_index = 0, nbook = 50, sort = nil, marker = nil)
      @cgi = cgi
      @library = library
      @start_index = start_index
      @nbook = nbook
      @sort = sort
      @accesskey = '.'
   end

   def url
      query_str(index_url,
		{:cmd => :cover_list,
		 :lib => @library.name,
		 :s => @start_index,
		 :n =>@nbook})
   end

   def body
      erb = ERB.new(<<-EOS, nil, '>')
    <div id="books">
<%       books_to_show.each do |book|	%>
<%          image_path = image_cache_url(book.image_url_medium)	%>
<%          vpage = ViewPage.new(@cgi, @library, book)	%>
      <span class="book">
        <a href="<%=h vpage.url %>" title="<%=h book.title %>"><%# %>
<%          unless image_path then	%>
<%#          %><span class="no-cover-book"><%=h book.title %></span><%# %>
<%          else	%>
<%#          %><img src="<%=h image_path %>" alt="<%=h book.title %>" /><%# %>
<%          end	%>
<%#    %></a>
      </span>
<%       end	%>
    </div>

<%= footer %>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def title
      _('Cover list of %s') % [@library.name] +
         if @nbook > 0 then
            _("(%d - %d / %d)") %
               [@start_index + 1,
                [@start_index + @nbook, @library.size].min,
                @library.size]
         else
            ''
         end
   end

   def navigation
      tp = TopPage.new(@cgi)
      sp = SearchPage.new(@cgi)
      lp = ListPage.new(@cgi, @library, @start_index)
      pp = self.prev_page
      np = self.next_page

      erb = ERB.new(<<-EOS, nil, '>')
      <div id="navigation">
	 <a href="<%=h tp.url %>" accesskey="<%=h tp.accesskey %>"><%=h _('top') %></a>
	 <a href="<%=h sp.url %>" accesskey="<%=h sp.accesskey %>"><%=h _('search') %></a>
	 <a href="<%=h lp.url %>" accesskey="<%=h lp.accesskey %>"><%=h _('list') %></a>
<% if pp then %>
	 <a href="<%=h pp.url %>" accesskey="<%=h pp.accesskey %>">
<% end %>
	     <%=h _('prev') %>
<% if pp then %>
	 </a>
<% end %>
<% if np then %>
	 <a href="<%=h np.url %>" accesskey="<%=h np.accesskey %>">
<% end %>
	     <%=h _('next') %>
<% if np then %>
	 </a>
<% end %>
      </div>
   <form id="books-per-page" action="<%=h index_url %>" method="get">
	 <label><select name="n">
<% if @nbook.zero? then %>
		    <option value="0" selected="selected"><%=h _('all') %></option>
<% else %>
		    <option selected="selected"><%=h @nbook %></option>
<% end %>
<% [15, 30, 50, 100].each do |n| %>
<%    unless n == @nbook then %>
		    <option><%=h n %></option>
<%    end %>
<% end %>
<% unless @nbook.zero? then %>
		    <option value="0"><%=h _('all') %></option>
<% end %>
		</select>
		<%=h _('books/page') %></label>
         <input type="hidden" name="lib" value="<%=h @library.name %>" />
         <input type="hidden" name="s" value="<%=h @start_index %>" />
         <input type="hidden" name="sort" value="<%=h @sort %>" />
         <input type="hidden" name="cmd" value="cover_list" />
         <input type="submit" value="<%=h _('show') %>" />
   </form>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end
end


class TopPage < HTMLPage
   def initialize(cgi)
      @cgi = cgi
   end

   def url
      query_str(index_url, {})
   end

   def body
      erb = ERB.new(<<-EOS, nil, '>')
   <form action="<%=h index_url %>" method="get">
      <div>
         <input type="text" name="id" value="" accesskey="0" />
         <input type="submit" value="<%=h _('View/Add') %>" />
         <input type="hidden" name="cmd" value="view" />
      </div>
      <div id="libraries">
<%       firstp = true	%>
<%       libpaths.each_with_index do |l, i|	%>
<%          lib = Cocoron::Library.new(l)	%>
<%          n = i + 1 %>
<%          idstr = 'lib-%d' % n %>
         <label accesskey="<%=h n %>">
	     <input type="radio" name="lib" value="<%=h lib.name %>" <%# %>
<%#                  %><%= 'checked="checked"' if firstp %> />
                <%=h lib.name %>(<%=h lib.size %>)</label>
<%       unless lib.empty? then	%>
         <a href="<%=h ListPage.new(@cgi, lib).url %>"><%=h _('[list]') %></a>
         <a href="<%=h StatPage.new(@cgi, lib).url %>"><%=h _('[stat]') %></a>
         <br />
<%       end	%>
<%          firstp = false	%>
<%       end	%>
      </div>
   </form>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def title
      Preference.instance['title']
   end

   def navigation
      sp = SearchPage.new(@cgi)
      erb = ERB.new(<<-EOS, nil, '>')
   <div id="navigation">
      <a href="<%=h sp.url %>" accesskey="<%=h sp.accesskey %>"><%=h _("search") %></a>
      <abbr title="RDF Site Summary">RSS</abbr>
      (<a href="<%=h RSSPage.new(@cgi, 'ctime').url %>"><%=h _("date added") %></a>/<a href="<%=h RSSPage.new(@cgi, 'mtime').url %>"><%=h _("last modified") %></a>)
   </div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def accesskey
      '0'
   end
end


class ViewPage < HTMLPage
   def initialize(cgi, library, book)
      @cgi = cgi
      @library = library
      @book = book
   end

   def url
      query_str(index_url,
                {:cmd => :view,
                 :lib => @library.name,
                 :id => @book.id.code})
   end

   def body
      cache_image!(@book)
      img_s = image_cache_url(@book.image_url_small)
      img_m = image_cache_url(@book.image_url_medium)
      img_l = image_cache_url(@book.image_url_large)
      @library[@book.id.code] = @book

      erb = ERB.new(<<-EOS, nil, '>')
   <div>
<%       if img_m then	%>
<%          if img_l then	%><a href="<%=h img_l %>"><%       end	%>
       <img id="img_m" src="<%=h img_m %>" alt="" />
<%          if img_l then	%></a><%       end	%>
<%	 elsif img_s then	%>
       <img id="img_s" src="<%=h img_s %>" alt="" />
<%       end	%>
   </div>
   <dl>
      <dt><%=h _(@book.id.sort) %></dt><dd><%=h @book.id.code %></dd>
      <dt><%=h _('title') %></dt><dd><%=h @book.title %></dd>
      <dt><%=h _('authors') %></dt><dd><%=h authors_to_s(@book.authors) %></dd>
      <dt><%=h _('publisher') %></dt><dd><%=h @book.publisher %></dd>
      <dt><%=h _('edition') %></dt><dd><%=h @book.edition %></dd>
      <dt><%=h _('release_date') %></dt><dd><%=h @book.release_date %></dd>
      <dt><%=h _('price') %></dt><dd><%=h @book.price %></dd>
      <dt><%=h _('date added') %></dt><dd><%=h @book.ctime.strftime("%Y/%m/%d") %></dd>
      <dt><%=h _('last modified') %></dt><dd><%=h @book.mtime.strftime("%Y/%m/%d") %></dd>
      <dt><%=h _('note') %></dt>
      <dd>
         <form method="post" action="<%=h update_url %>">
            <div><textarea name="note" cols="80" rows="5" accesskey="-"><%# %>
<%#                  %><%=h @book.note %></textarea></div>
            <div>
               <input type="submit" accesskey="=" <%# %>
<%#                %> value="<%=h _("write note") %>" />
               <input type="hidden" name="id" value="<%=h @book.id.code %>" />
               <input type="hidden" name="lib" value="<%=h @library.name %>" />
               <input type="hidden" name="cmd" value="write_note" />
            </div>
         </form>
      </dd>
   </dl>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def title
      @book.title
   end

   def navigation
      amazon_addr = 'http://www.amazon.co.jp/exec/obidos/ASIN/'

      tp = TopPage.new(@cgi)
      ep = EditPage.new(@cgi, @library, @book)
      sp = SearchPage.new(@cgi)
      lp = ListPage.contains(@book, @library, @cgi)

      erb = ERB.new(<<-EOS, nil, '>')
   <div id="navigation">
      <a href="<%=h tp.url %>" accesskey="<%=h tp.accesskey %>"><%=h _('top') %></a>
      <a href="<%=h ep.url %>" accesskey="<%=h ep.accesskey %>"><%=h _('edit') %></a>
      <a href="<%=h sp.url %>" accesskey="<%=h sp.accesskey %>"><%=h _('search') %></a>
      <a href="<%=h lp.url %>" accesskey="<%=h lp.accesskey %>"><%=h _('list') %></a>
<% if @book.id.sort == 'ISBN' then %>
      <a href="<%=h amazon_addr + Cocoron::BookID::ISBN.isbn13to10(@book.id.code) %>">Amazon</a>
<% end %>
   </div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end
end


class ErrorPage < HTMLPage
   def initialize(cgi, err_msg)
      @cgi = cgi
      @msg = err_msg
   end

   def title
      'Cocoron::Error'
   end

   def body
      <<-EOS
      <pre>#{CGI.escapeHTML(@msg)}</pre>
      EOS
   end

   def navigation
      tp = TopPage.new(@cgi)

      erb = ERB.new(<<-EOS, nil, '>')
      <div id="navigation"><a href="<%=h tp.url %>" accesskey="<%=h tp.accesskey %>">top</a></div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end
end


class SearchPage < HTMLPage
   def initialize(cgi)
      @cgi = cgi
   end

   def self.compile_query(cgi)
      attrs_used_to_search.inject(Hash.new) do |result, k|
         if cgi.params[k].first and not cgi.params[k].first.empty? then
            v = cgi.get_param(k)
            result[k] = case k
			when 'authors' then
			   p = Regexp.new(Regexp.quote(v), Regexp::IGNORECASE)
			   lambda{|as| not as.nil? and as.any?{|a| p =~ a }}
			when 'release_date' then
			   p = Regexp.new("^#{Regexp.quote(v)}")
			   lambda{|s| p =~ s }
			when 'mtime', 'ctime' then
			   p = Regexp.new("^#{Regexp.quote(v)}")
			   lambda{|time|
			      p =~ time.strftime("%Y/%m/%d") }
			when 'note' then
			   ps = v.split(/\s+/o).collect{|s|
			           Regexp.new(Regexp.quote(s)) }
			   lambda{|s| ps.any?{|p| p =~ s }}
			when 'price' then
			   ps = v.split(/\s+/o).collect{|s|
				   if s.sub!(/^((?:<|>)=?|==)/o, '') then
				      cmp = $1
				      lambda{|t|
					 t.delete('^0-9').to_i\
					    .send(cmp, s.delete('^0-9').to_i) }
				   else
				      lambda{|t|
					 s.delete('^0-9') == t.delete('^0-9') }
				   end }
			   lambda{|price| ps.all?{|p| p.call(price) }}
			else
			   p = Regexp.new(Regexp.quote(v), Regexp::IGNORECASE)
			   lambda{|s| p =~ s }
			end
         end

	 result
      end
   end

   def body
      erb = ERB.new(<<-EOS, nil, '>')
   <form action="<%=h index_url %>" method="get">
      <ul id="query">
         <li><label><%=h _('title') %> 
             <input type="text" name="title" size="40"
                 value="<%=h @cgi.get_param('title'){''} %>" /></label></li>
         <li><label><%=h _('authors') %> 
             <input type="text" name="authors" size="40"
                 value="<%=h @cgi.get_param('authors'){''} %>" /></label></li>
         <li><label><%=h _('publisher') %> 
             <input type="text" name="publisher" size="40"
                 value="<%=h @cgi.get_param('publisher'){''} %>" /></label></li>
         <li><label><%=h _('edition') %> 
             <input type="text" name="edition" size="40"
                 value="<%=h @cgi.get_param('edition'){''} %>" /></label></li>
         <li><label><%=h _('release_date') %> 
             <input type="text" name="release_date" size="40"
                  value="<%=h @cgi.get_param('release_date'){''} %>" /></label></li>
         <li><label><%=h _('price') %> 
             <input type="text" name="price" size="40"
                  value="<%=h @cgi.get_param('price'){''} %>" /></label></li>
         <li><label><%=h _('note') %> 
             <input type="text" name="note" size="40"
                  value="<%=h @cgi.get_param('note'){''} %>" /></label></li>
      </ul>

      <div id="libraries">
<%       i = 0 %>
<%       libpaths.each_slice(4) do |lps|	%>
         <div>
<%          lps.each do |lp|	%>
<%             lib = Cocoron::Library.new(lp)	%>
            <label accesskey="<%=h i + 1 %>">
		<input type="checkbox" name="libs"<%# %>
<%= @cgi.get_params('libs').include?(lib.name) ? 'checked="checked" ' : '' %>
<%#                    %> value="<%=h lib.name %>" /> <%=h lib.name %></label>
<%             i += 1	%>
<%          end	%>
         </div>
<%       end	%>
      </div>

      <div id="submit_and_reset">
         <input type="hidden" name="cmd" value="search" />
         <input type="submit" name="submit" value="<%=h _('search') %>" />
         <input type="reset" value="<%=h _('reset') %>" id="reset" />
      </div>
   </form>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def title
      Preference.instance['title'] + _('::') + _('Search')
   end

   def navigation
      tp = TopPage.new(@cgi)
      erb = ERB.new(<<-EOS, nil, '>')
   <div id="navigation">
      <a href="<%=h tp.url %>" accesskey="<%=h tp.accesskey %>"><%=h _('top') %></a>
   </div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def accesskey
      '/'
   end

   def url
      query_str(index_url, {:cmd => :search})
   end

private
   def self.attrs_used_to_search
      %w(title authors publisher edition release_date price note mtime ctime)
   end
end


class SearchResultPage < HTMLPage
   def initialize(cgi, libs, crits)
      @cgi = cgi
      @libs = libs
      @crits = crits
   end

   def body
      books = []

      @libs.each do |libname|
         library = load_library(libname)

         library.each_value do |book|
            if @crits.all? {|k, pred| pred.call(book.send(k)) } then
               books << [library, book]
            end
         end
      end

      erb = ERB.new(<<-EOS, nil, '>')
<%       if books.empty? then	%>
   <p><%=h _('No book is found.') %></p>
<%       else	%>
   <p><%=h n_('Only %d book is found.', '%d books are found.', books.size) %
              [books.size] %></p>
   <table border="1" summary="<%=h self.title %>">
   <thead>
      <tr>
         <th><%=h _('library') %></th>
         <th><%=h _('ID') %></th>
         <th><%=h _('title') %></th>
         <th><%=h _('authors') %></th>
         <th><%=h _('publisher') %></th>
         <th><%=h _('edition') %></th>
         <th><%=h _('release_date') %></th>
         <th><%=h _('price') %></th>
         <th><%=h _('note') %></th>
      </tr>
   </thead>
   <tbody>
<%          books.each do |library, book|	%>
<%             vpage = ViewPage.new(@cgi, library, book)	%>
      <tr>
         <td><%=h library.name %></td>
         <td><a href="<%=h vpage.url %>"><%=h book.id.code %></a></td>
         <td><%=h book.title %></td>
         <td><%=h authors_to_s(book.authors) %></td>
         <td><%=h book.publisher %></td>
         <td><%=h book.edition %></td>
         <td><%=h book.release_date %></td>
         <td><%=h book.price %></td>
         <td><pre><%=h book.note %></pre></td>
      </tr>
<%          end	%>
   </tbody>
   </table>
<%       end	%>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def title
      Preference.instance['title'] + _('::') + _('Search Result')
   end

   def navigation
      tp = TopPage.new(@cgi)
      sp = SearchPage.new(@cgi)

      erb = ERB.new(<<-EOS, nil, '>')
   <div id="navigation">
      <a href="<%=h tp.url %>" accesskey="<%=h tp.accesskey %>"><%=h _('top') %></a>
      <a href="<%=h sp.url %>" accesskey="<%=h sp.accesskey %>"><%=h _('search') %></a>
   </div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end
end


class MoveConfirmationPage < HTMLPage
   def initialize(cgi, book, from, to)
      @cgi = cgi
      @book = book
      @from = from
      @to = to
   end

   def title
      Preference.instance['title']
   end

   def body
      erb = ERB.new(<<-EOS, nil, '>')
   <p><%=h _("Really move `%s'(%s) from %s to %s?") %
              [ @book.title, @book.id.to_s, @from.name, @to.name] %></p>
   <div id="answers">
      <form action="<%=h index_url %>" method="post">
         <input type="submit" id="cancel" value="<%=h _('Cancel') %>" />
         <input type="hidden" name="cmd" value="edit" />
         <input type="hidden" name="id" value="<%=h @book.id.code %>" />
         <input type="hidden" name="lib" value="<%=h @from.name %>" />
      </form>
      <form action="<%=h update_url %>" method="post">
         <input type="submit" id="do_it" value="<%=h _('Move') %>" />
         <input type="hidden" name="cmd" value="move" />
         <input type="hidden" name="id" value="<%=h @book.id.code %>" />
         <input type="hidden" name="from" value="<%=h @from.name %>" />
         <input type="hidden" name="to" value="<%=h @to.name %>" />
      </form>
   </div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end
end


class RemovalConfirmationPage < HTMLPage
   def initialize(cgi, library, book)
      @cgi = cgi
      @library = library
      @book = book
   end

   def title
      Preference.instance['title']
   end

   def body
      erb = ERB.new(<<-EOS, nil, '>')
   <p><%=h _("Really remove `%s'(%s) from %s?") %
              [ @book.title, @book.id.to_s, @library.name] %></p>
   <div id="answers">
      <form action="<%=h index_url %>" method="post">
         <input type="submit" id="cancel" value="<%=h _('Cancel') %>" />
         <input type="hidden" name="cmd" value="edit" />
         <input type="hidden" name="id" value="<%=h @book.id.code %>" />
         <input type="hidden" name="lib" value="<%=h @library.name %>" />
      </form>
      <form action="<%=h update_url %>" method="post">
         <input type="submit" id="do_it" value="<%=h _('Remove') %>" />
         <input type="hidden" name="cmd" value="remove" />
         <input type="hidden" name="id" value="<%=h @book.id.code %>" />
         <input type="hidden" name="lib" value="<%=h @library.name %>" />
      </form>
   </div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end
end


class EditPage < HTMLPage
   def initialize(cgi, library, book)
      @cgi = cgi
      @library = library
      @book = book
   end

   def url
      query_str(index_url,
                {:cmd => :edit,
                 :id => @book.id.code,
                 :lib => @library.name})
   end

   def body
      erb = ERB.new(<<-EOS, nil, '>')
<%       unless @library.has_key?(@book.id.code) then	%>
    <p><%=h _('%s does not exist in %s.') % [@book.id.to_s, @library.name] %>
    </p>
<%       else	%>
   <form action="<%=h index_url %>" method="post">
      <div>
         <input type="hidden" name="id" value="<%=h @book.id.code %>" />
         <input type="hidden" name="lib" value="<%=h @library.name %>" />
         <input type="hidden" name="cmd" value="remove" />
         <input type="submit" value="<%=h _('remove') %>" />
      </div>
   </form>
   <form action="<%=h index_url %>" method="post">
      <p>
         <input type="hidden" name="id" value="<%=h @book.id.code %>" />
         <input type="hidden" name="from" value="<%=h @library.name %>" />
         <input type="hidden" name="cmd" value="move" />
         <label><%=h _('library') %>:
            <select name="to">
<%          libpaths.each do |path|	%>
<%             lib = Cocoron::Library.new(path)	%>
               <option<%# %>
<%#           %><%= lib.name == @library.name ?
                       ' selected="selected" disabled="disabled"' :
                       ''%>><%=h lib.name %></option>
<%          end	%>
            </select>
         </label>
         <input type="submit" value="<%=h _('move') %>" />
      </p>
   </form>
<%       end	%>
   <form action="<%=h update_url %>" method="post">
      <ul>
         <li><label><%=h _('title') %>
<%#       %> <input type="text" name="title"<%# %>
<%#             %> value="<%=h @book.title %>" size="40" />
             </label></li>
         <li><label><%=h _('authors') %> 
             <textarea name="authors" cols="40" rows="5"><%# %>
<%#           %><%=h(@book.authors ? @book.authors.join("\n") : '') %>
<%#       %></textarea>
             </label></li>
         <li><label><%=h _('publisher') %>
<%#       %> <input type="text" name="publisher"<%# %>
<%#             %> value="<%=h @book.publisher %>" size="40" />
             </label></li>
         <li><label><%=h _('edition') %>
<%#       %> <input type="text" name="edition"<%# %>
<%#             %> value="<%=h @book.edition %>" size="40" />
             </label></li>
         <li><label><%=h _('release_date') %>
<%#       %> <input type="text" name="release_date"<%# %>
<%#             %> value="<%=h @book.release_date %>" size="40" />
             </label></li>
         <li><label><%=h _('price') %>
<%#       %> <input type="text" name="price"<%# %>
<%#             %> value="<%=h @book.price %>" size="40" />
             </label></li>
         <li><label><%=h _('image_url_small') %>
<%#       %> <input type="text" name="image_url_small"<%# %>
<%#             %> value="<%=h @book.image_url_small %>" size="40" />
             </label></li>
         <li><label><%=h _('image_url_medium') %>
<%#       %> <input type="text" name="image_url_medium"<%# %>
<%#             %> value="<%=h @book.image_url_medium %>" size="40" />
             </label></li>
         <li><label><%=h _('image_url_large') %>
<%#       %> <input type="text" name="image_url_large"<%# %>
<%#             %> value="<%=h @book.image_url_large %>" size="40" />
             </label></li>
      </ul>
      <div>
         <input type="hidden" name="id" value="<%=h @book.id.code %>" />
         <input type="hidden" name="lib" value="<%=h @library.name %>" />
         <input type="hidden" name="cmd" value="commit" />
         <input type="submit" value="<%=h _('commit') %>" />
      </div>
   </form>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def title
      if @library.has_key?(@book.id.code) then
         _('Edit %s') % [@book.id.to_s]
      else
         _('Add %s') % [@book.id.to_s]
      end
   end

   def navigation
      tp = TopPage.new(@cgi)
      lp = ListPage.contains(@book, @library, @cgi)

      erb = ERB.new(<<-EOS, nil, '>')
   <div id="navigation">
      <a href="<%=h tp.url %>" accesskey="<%=h tp.accesskey %>"><%=h _('top') %></a>
      <a href="<%=h lp.url %>" accesskey="<%=h lp.accesskey %>"><%# %>
<%=h _('list') %></a>
   </div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def accesskey
      ':'
   end
end



class StatPage < HTMLPage
   def initialize(cgi, library)
      @cgi = cgi
      @library = library
   end

   def url
      query_str(index_url,
		{:cmd => :stat,
		 :lib => @library.name})
   end

   def title
      _('Statistical information of %s') % @library.name
   end

   def body
      stat, total_price = calc_stat
      nbooks = @library.size

      erb = ERB.new(<<-EOS, nil, '>')
    <p><%=h _('# of books') %>: <%=h nbooks %></p>

    <h2><%=h _('authors') %></h2>

    <table border="1" id="author">
        <thead>
            <tr>
               <th><%=h _('#') %></th>
               <th><%=h _('authors') %></th>
               <th><%=h _('ratio') %></th>
            </tr>
         </thead>
         <tbody>
<%       calc_rank(stat[:author]).each do |rank, (author, n)|	%>
             <tr>
                 <th><%=h rank %></th>
                 <th><a href="<%=h search_url(:authors, author) %>"><%# %>
<%#                %><%=h author %></a></th>
                 <td>
                     <img src="images/bar.gif" height="15" <%# %>
<%#                     %>width="<%=h (hist_max_width * n / nbooks).ceil %>">
                     <%=h n %> / <%=h '%3.1f' % (100.0 * n / nbooks)  %> %
                 </td>
             </tr>
<%       end	%>
         </tbody>
    </table>

    <h2><%=h _('publisher') %></h2>

    <table border="1" id="publisher">
        <thead>
            <tr>
               <th><%=h _('#') %></th>
               <th><%=h _('publisher') %></th>
               <th><%=h _('ratio') %></th>
            </tr>
         </thead>
         <tbody>
<%       calc_rank(stat[:publisher]).each do |rank, (publisher, n)|	%>
             <tr>
                 <th><%=h rank %></th>
                 <th><a href="<%=h search_url(:publisher, publisher) %>"><%# %>
                     <%=h publisher %></a></th>
                 <td>
                     <img src="images/bar.gif" height="10" <%# %>
<%#                     %>width="<%=h (hist_max_width * n / nbooks).ceil %>">
                     <%=h n %> / <%=h '%3.1f' % (100.0 * n / nbooks)  %> %
                 </td>
             </tr>
<%       end	%>
         </tbody>
    </table>

    <h2><%=h _('release_date') %></h2>

    <table border="1" id="release_date">
        <thead>
            <tr>
               <th><%=h _('release_date') %></th>
               <th><%=h _('ratio') %></th>
            </tr>
         </thead>
         <tbody>
<%       stat[:release_date]\
            .sort_by{|(d, n)| d }.reverse_each do |(release_date, n)|	%>
             <tr>
                 <th><a href="<%=h search_url(:release_date, release_date) %>"><%=h release_date %></a></th>
                 <td>
                     <img src="images/bar.gif" height="10" <%# %>
<%#                     %>width="<%=h (hist_max_width * n / nbooks).ceil %>">
                     <%=h n %> / <%=h '%3.1f' % (100.0 * n / nbooks)  %> %
                 </td>
             </tr>
<%       end	%>
         </tbody>
    </table>

    <h2><%=h _('date added') %></h2>

    <table border="1" id="date-of-entry">
        <thead>
            <tr>
               <th><%=h _('date added') %></th>
               <th><%=h _('ratio') %></th>
            </tr>
         </thead>
         <tbody>
<%       stat[:ctime]\
            .sort_by{|(d, n)| d }.reverse_each do |(ctime, n)|	%>
             <tr>
                 <th><a href="<%=h search_url(:ctime, ctime) %>"><%=h ctime %></a></th>
                 <td>
                     <img src="images/bar.gif" height="10" <%# %>
<%#                     %>width="<%=h (hist_max_width * n / nbooks).ceil %>">
                     <%=h n %> / <%=h '%3.1f' % (100.0 * n / nbooks)  %> %
                 </td>
             </tr>
<%       end	%>
         </tbody>
    </table>

    <h2><%=h _('last modified') %></h2>

    <table border="1" id="last-modified">
        <thead>
            <tr>
               <th><%=h _('last modified') %></th>
               <th><%=h _('ratio') %></th>
            </tr>
         </thead>
         <tbody>
<%       stat[:mtime]\
            .sort_by{|(d, n)| d }.reverse_each do |(mtime, n)|	%>
             <tr>
                 <th><a href="<%=h search_url(:mtime, mtime) %>"><%=h mtime %></a></th>
                 <td>
                     <img src="images/bar.gif" height="10" <%# %>
<%#                     %>width="<%=h (hist_max_width * n / nbooks).ceil %>">
                     <%=h n %> / <%=h '%3.1f' % (100.0 * n / nbooks)  %> %
                 </td>
             </tr>
<%       end	%>
         </tbody>
    </table>

<%       unless stat[:note].empty? then %>
    <h2><%=h _('note') %></h2>
    <ul>
<%          stat[:note].sort_by{|(note, n)| n }.reverse_each do |(note, n)| %>
        <li><a href="<%=h search_url(:note, note) %>"><%=h note %></a>
        (<%=h n %>)</li>
<%          end %>
    </ul>
<%       end %>

    <h2><%=h _('price') %></h2>

    <table border="1" id="price">
        <thead>
            <tr>
               <th><%=h _('price') %></th>
               <th><%=h _('ratio') %></th>
            </tr>
         </thead>
         <tbody>
<%       stat[:price].sort_by{|(p, n)| p }.each do |(price, n)|	%>
<%           lb = price_step * price	%>
<%           ub = lb + price_step	%>
<%# %>
             <tr>
                 <th><a href="<%=h search_url(:price, ">=%s <%s" % [lb, ub]) %>"><%=h lb %>-<%=h ub %></a></th>
                 <td>
                     <img src="images/bar.gif" height="10" <%# %>
<%#                     %>width="<%=h (hist_max_width * n / nbooks).ceil %>">
                     <%=h n %> / <%=h '%3.1f' % (100.0 * n / nbooks)  %> %
                 </td>
             </tr>
<%       end	%>
         </tbody>
    </table>

    <p><%=h _('average price') %>: <%=h commify(total_price / nbooks) %></p>
    <p><%=h _('total price') %>: <%=h commify(total_price) %></p>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

   def navigation
      tp = TopPage.new(@cgi)

      erb = ERB.new(<<-EOS, nil, '>')
   <div id="navigation">
      <a href="<%=h tp.url %>" accesskey="<%=h tp.accesskey %>"><%=h _('top') %></a>
   </div>
      EOS

      erb.filename = __FILE__
      erb.result(binding)
   end

private
   def hist_max_width
      320.0
   end

   def calc_stat
      stat = Hash.new{|h, k| h[k] = Hash.new(0) }
      total_price = 0

      @library.each do |_, book|
	 if book.authors then
	    book.authors.each do |a|
	       stat[:author][a] += 1
	    end
	 end

         if book.note and not book.note.empty? then
            book.note.each do |line|
               stat[:note][line] += 1
            end
         end

	 stat[:publisher][book.publisher] += 1
	 stat[:release_date][process_date(book.release_date)] += 1
	 stat[:price][process_price(book.price)] += 1
	 stat[:ctime][process_time(book.ctime)] += 1
	 stat[:mtime][process_time(book.mtime)] += 1

	 total_price += book.price.delete('^0-9').to_i
      end

      [stat, total_price]
   end

   def calc_rank(xns)
      xns.group_by{|(x, n)| n }\
	 .sort_by{|x_ns| x_ns.first.last }\
         .reverse!\
         .inject([1, Array.new]){|(rank, result), xs|
	 [rank + xs.length, result.concat(xs.sort.collect{|x| [rank, x] })]
      }.last
   end

   def process_date(str)
      (not str or str.empty?) ? '' : str[0,4]
   end

   def process_price(str)
      str.delete('^0-9').to_i.div(price_step)
   end

   def process_time(t)
      t.strftime('%Y/%m')
   end

   def price_step
      500
   end

   def commify(n)
      n.to_s.gsub(/(\d)(?=(?:\d{3})+(?!\d))/o, '\1,')
   end

   def search_url(attr, criterion)
      query_str(index_url,
		{:cmd => :search,
		 :libs => @library.name,
		 attr => criterion})
   end
end
