#!/usr/bin/env ruby

require 'delegate'
require 'fileutils'
require 'pstore'
require 'set'
require 'yaml'


module Cocoron
   class LibraryImplementation
      attr_reader :path

      def initialize
	 raise NotImplementedError
      end

      def [](id_str)
	 raise NotImplementedError
      end

      def []=(id_str, book)
	 raise NotImplementedError
      end

      def delete(id_str)
	 raise NotImplementedError
      end

      def inspect
	 "#<#{self.class} path=#{self.path}>"
      end

      def name
	 File.basename(@path)
      end

      def writable?
	 @writablep
      end

      def clear
	 self.each_keys do |key|
	    self.delete(key)
	 end
      end

      def each(&block)
	 each_keys do |key|
	    yield(key, self[key])
	 end
      end

      alias :each_pair :each

      def each_keys(&block)
	 keys.each(&block)
      end

      def each_value
	 each do |k, v|
	    yield(v)
	 end
      end

      def delete_if(&block)
	 self.each do |key, val|
	    self.delete(key) if yield(key, val)
	 end
      end

      def empty?
	 self.keys.empty?
      end

      def has_key?(key)
	 self.keys.include?(key)
      end

      alias :key? :has_key?
      alias :include? :has_key?

      def has_value?(val)
	 self.each_value do |v|
	    return true if v == val
	 end

	 return false
      end

      alias :value? :has_value?

      def length
	 self.keys.length
      end

      alias :size :length

      def values
	 result = []

	 self.each_value do |val|
	    result << val
	 end

	 result
      end

      def values_at(*keys)
	 keys.collect{|key| self[key] }
      end

   private
      def assert_is_writable
	 raise TypeError, "Library is not writable!" unless self.writable?
      end

      def assert_is_book(obj)
	 raise TypeError, "#{obj} is not a Book!" unless obj.is_a?(Book)
      end
   end

   class DLibrary < LibraryImplementation
      DB_ENTRY_EXT = '.yaml'

      def initialize(path, mode = :r)
	 if mode == :r and not File.exist?(path) then
	    raise ArgumentError, "No such Library: #{path}"
	 end

	 if mode == :w and not File.writable?(path) then
	    raise ArgumentError, "Library is not writable: #{path}"
	 end

	 @path = path
	 @cache = Hash.new
	 @writablep = (mode == :w)
      end

      def [](id_str)
	 epath = entry_path(id_str)

	 if @cache[id_str] then
	    @cache[id_str]
	 elsif File.exist?(epath) then
	    result = File.open(epath){|dbent| YAML.load(dbent) }

	    raise TypeError, "#{result} is not a Book" \
	       unless result.is_a?(Book)

	    @cache[id_str] = result
	    result
	 else
	    nil
	 end
      end

      def []=(id_str, book)
	 assert_is_book(book)
	 assert_is_writable

	 @cache[id_str] = book

	 open_with_lock(entry_path(id_str), 'w') do |f|
	    f.write book.to_yaml
	 end

	 book
      end

      def delete(id_str)
	 assert_is_writable

	 if self.has_key?(id_str) then
	    File.delete(entry_path(id_str))
	    @cache.delete(id_str)
	 elsif block_given? then
	    yield(id_str)
	 else
	    nil
	 end
      end

      def keys
	 Dir.glob(entry_path('[0-9]*')).collect{|fn|
	    File.basename(fn, DB_ENTRY_EXT).untaint
	 }.sort
      end

   private
      def open_with_lock(path, mode = 'r', perm = 0666, max_retry = 10, timeout_sec = 600, lock_ext = '.lock', &block)
	 n = 0
	 lock_name = path + lock_ext

	 begin
	    Dir.mkdir(lock_name)
	 rescue SystemCallError
	    if n >= max_retry then
	       raise "Cannot open #{path} exclusively"
	    elsif Time.now.to_i - File.mtime(lock_name).to_i >= timeout_sec then
	       Dir.rmdir(lock_name)
	    else
	       n += 1
	       sleep 1
	    end

	    retry
	 end

	 begin
	    open(path, mode, perm){|f| yield(f) }
	 ensure
	    Dir.rmdir(lock_name)
	 end
      end

      def entry_path(key)
	 File.join(@path, key + DB_ENTRY_EXT)
      end
   end

   class FLibrary < LibraryImplementation
      def initialize(path, mode = :r)
	 @writablep = (mode == :w)
	 @path = path
	 @modified = Set.new
	 @entries = PStore.new(path).transaction(true){|db|
		       db.roots.inject(Hash.new){|h, k|
			  h[k] = db[k]; h
		       }
		    }

	 @finalizer = FLibrary.make_finalizer(@path, @entries, @modified)

	 ObjectSpace.define_finalizer(self, @finalizer)
      end

      def [](id_str)
	 @entries[id_str]
      end

      def []=(id_str, book)
	 assert_is_writable
	 assert_is_book(book)

	 @entries[id_str] = book

	 @modified.add(id_str)
      end

      def delete(id_str, &block)
	 assert_is_writable

	 @entries.delete(id_str, &block)

	 @modified.add(id_str)
      end

      def keys
	 @entries.keys.sort
      end

      def self.make_finalizer(path, entries, modified)
	 lambda{
	    unless modified.empty? then
	       PStore.new(path).transaction do |db|
		  modified.each do |id|
		     e = entries[id]

		     if e.nil? then
			db.delete(id)
		     else
			db[id] = e
		     end
		  end
	       end
	    end
	 }
      end
   end

   class Library < DelegateClass(LibraryImplementation)
      include Enumerable

      def initialize(path, mode = :r)
	 unless [:r, :w].include?(mode) then
	    raise ArgumentError, "Unknown mode: #{mode}"
	 end

	 super((File.directory?(path) ? DLibrary : FLibrary).new(path, mode))
      end
   end
end
